#include "aspps.h"

int LookAHead(int *literal);


struct BackTrack *NewStck(void)
{
	struct BackTrack *tmp;

	if((tmp=(struct BackTrack *)malloc(sizeof(struct BackTrack)))==NULL){
		fprintf(Errorfile,"out of memory\n");
		exit(1);
	}
	memset(tmp,0,sizeof(struct BackTrack));
	return(tmp);
}

int *NewData(size_t number,int val)
{
	int *tmp;
	  
	if(!number)
		return(NULL);
	if((tmp=(int *)malloc(number))==NULL){
		fprintf(Errorfile,"cannot allocate Clause list\n");
		exit(1);
	}
   tmp=memset(tmp,val,number);
	return(tmp);
}


void ParseTxtVar(char *input)
{
	int		i,len;
	int		index;
	char	buffer[1000000];

	len=strlen(input)-1;
	i=0;
	while(i<len&&!isdigit((int)input[i])) i++;
	index=atoi(&input[i]);
	while(i<len&&isdigit((int)input[i])) i++;
	input[len]='\0';
	while(i<len&&isspace((int)input[i])) i++;
	
	if(index)
		Atoms[index].txt= strdup(&input[i]);
	else {
		sprintf(buffer,"%s\n%s",Atoms[index].txt,&input[i]); 
		free(Atoms[index].txt);
		Atoms[index].txt=strdup(buffer);
	}
}

int ParseInt(char *input, int *begin)
{
	int i,j;
	char line[180];

	i=*begin;
	while(input[i]!=EOLN&&isspace((int)input[i]))i++;
	if(input[i]==EOLN || input[i]=='}')
		return(-1);
	j=0;
	while(isdigit((int)input[i])) line[j++]=input[i++];
	line[j]=EOLN;
	(*begin)=i;
	return(atoi(line));
}

int *GetList(char *input,int *index,int *count)
{
	int		i,j,size;
	int	*tmp=NULL;

	size=0;
	i=(*index);
	while((ParseInt(input,&i)>=0)) size++;
	if(size){
		tmp=(int *)NewData(sizeof(int)*size,0);
		(*count)=size;
		i=(*index);
		for(j=0;j<size;j++){
			tmp[j]=ParseInt(input,&i);
		}
	}
	(*index)=i;
	return(tmp);
}

int ParseAtom(char *input,int *index)
{
	int	   count=0,pos,neg,*list;
	char	buffer[180];
	

	(*index)++;
	pos=ParseInt(input,index);
	neg=ParseInt(input,index);
	list=GetList(input,index,&count);
	if(Debug){
		if(list)
			fprintf(stderr,"pos=%i neg=%i list[0]=%i count=%i\n",
					pos,neg,list[0],count);
		else
			fprintf(stderr,"pos=%i neg=%i list=NULL count=%i\n",pos,neg,count);
	}
	if(Maxlit<count) Maxlit=count;
	Meta++;
	sprintf(buffer,"Meta%i",Meta);
	Atoms[Meta].txt= strdup(buffer);
	Atoms[Meta].meta=1;
	Atoms[Meta].pos=pos;
	if(count-neg<0)
		Atoms[Meta].neg=0;
	else Atoms[Meta].neg=count-neg;
	Atoms[Meta].metalist=list;
	Atoms[Meta].cnt=count;
	return(Meta);
}


int ParseRule(char *input,int index)
{
	int		i=0,count=0,hd=0;
	
	
	while(i<strlen(input)){
		if(input[i]=='{'){
			if(hd)
				Rules[index].hdcnt++;
			else Rules[index].bdycnt++;
			while(input[i]!='}') i++;
		}
		else if(input[i]==',')
			hd=1;
		else if(input[i]==':'){
			hd=1;
			Rules[index].horn=1;
		}
		else if(isdigit((int)input[i])){
			if(hd)Rules[index].hdcnt++;
			else Rules[index].bdycnt++;
			while(isdigit((int)input[i]))i++;
		}
		i++;
	}
	if(Rules[index].hdcnt+Rules[index].bdycnt==0) return(0);
	Rules[index].hdptr=(int *)NewData(sizeof(int)*Rules[index].hdcnt,0);
	Rules[index].bdyptr=(int *)NewData(sizeof(int)*Rules[index].bdycnt,0);
	Rules[index].hdcnt=Rules[index].bdycnt=i=hd=0;
	while(i<strlen(input)){
		if(input[i]==','||input[i]==':')
			hd=1;
		else if(input[i]=='{'){
			if(hd)
				Rules[index].hdptr[Rules[index].hdcnt++]=ParseAtom(input,&i);
			else
				Rules[index].bdyptr[Rules[index].bdycnt++]=ParseAtom(input,&i);
		}
		else if(isdigit((int)input[i])){
			if(hd)
				Rules[index].hdptr[Rules[index].hdcnt++]=ParseInt(input,&i);
			else
				Rules[index].bdyptr[Rules[index].bdycnt++]=ParseInt(input,&i);
		}

		i++;
	}

	if(!Rules[index].horn)
		Nsat++;
	count=Rules[index].hdcnt+Rules[index].bdycnt;
	if(Maxlit<count)
		Maxlit=count;
	else if(Minlit>count)
		Minlit=count;
	return(1);
}
		
void SetCnt(void)
{
	int i,increase=1;;

	for(i=1;i<=Outvar;i++)
		if(AList[i].val==-1)
			increase*=2;
	Cnt+=increase;
}

void PrintAnswer(void)
{
	int	i,j,k;
	char	buffer[180];

	if(Cnt==0)
		fprintf(stdout,"\nAtoms determined during grounding:%s\n",Atoms[0].txt);
	for(j=0;j<ShowCnt;j++)
		printf(" %s\n",ShowAtoms[j]);
	fprintf(stdout,"\nAnswer-set %i:\n",Cnt+1);
	for(i=1;i<=Outvar;i++)
		if(AList[i].val!=0){
			if(!ShowCnt){
				if(AList[i].val==-1){
					printf(" not %s or",Atoms[i].txt);
				}
				printf(" %s\n",Atoms[i].txt);
			}
			else{
				j=0;
				while(j<strlen(Atoms[i].txt)&&(isdigit(Atoms[i].txt[j])||isspace(Atoms[i].txt[j])))
					j++;
				k=0;
				buffer[k]='\0';
				while(j<strlen(Atoms[i].txt)&&Atoms[i].txt[j]!='(')
					buffer[k++]=Atoms[i].txt[j++];
				buffer[k]='\0';
				j=0;
				while(j<ShowCnt){
					if(!strcmp(buffer,ShowAtoms[j])){
						if(AList[i].val!=1)
							printf(" not %s or",Atoms[i].txt);
						printf(" %s\n",Atoms[i].txt);
					}
					j++;
				}
			}
		}
	fprintf(stdout,"\n");
}


													   
int GetProblem(void)
{
	int		index,i;
	char	line[1800];


	index=0;
	Nsat=0;
	while (fgets(line,1799,fptr)&&line[0]!='p');
	if(line[0]=='p'){
		i=1; 
		Outvar = Meta =ParseInt(line,&i);
		if(Debug){
			fprintf(Errorfile, "Outvar = % i\n", Outvar);
		}
		i++;
		Nvar= Outvar + ParseInt(line,&i);
		if(Debug){
			fprintf(Errorfile, "Nvar = % i\n", Nvar);
		}
		i++;
		Nrule=ParseInt(line,&i);
		if(Debug){
			fprintf(Errorfile, "Nrule = % i\n", Nrule);
		}
	}
	else{
		printf("error: in format of line 1 of input \n");
		return(-1);
	}
	Atoms=(struct Atom *)NewData(sizeof(struct Atom)*(Nvar+1),0);
	Atoms[0].txt=strdup("");
	Rules=(struct Rule *)NewData(sizeof(struct Rule)*Nrule,0);
	while (fgets(line,1799,fptr)){
		switch(line[0]){
			case 'p': break;
			case 'c': ParseTxtVar(line); break;
			default: 
				if(!ParseRule(line,index++)) return(0); 
				break;
		}
	}
	if(fptr != stdin)
		fclose(fptr);
	if(Nrule!=index||Meta!=Nvar){
		printf("Input Error:  Nrule=%i index=%i -- Meta=%i Nvar=%i\n",
				Nrule,index,Meta,Nvar);
		return(-1);
	}
	SetAtomQueue=NewData(sizeof(int)*Nvar,0);
	AtomQueueVal=(short *)NewData(sizeof(short)*(Nvar+1),-1);
	PathList=(short *)NewData(sizeof(short)*(Nvar+1),0);
	Tmplist=(double *)NewData(sizeof(double)*(Nvar+1),0);
	if(Debug){
		fprintf(Errorfile,
			"Maxlit = %i Minlit = %i Nsat= %i Nrule = %i Nvar = % i\n",
			Maxlit,Minlit,Nsat,Nrule,Nvar);
	}
	return(1);
}

struct RuleStk *SetCurrentRules(void)
{
	struct RuleStk *tmp;
	
	if((tmp=(struct RuleStk *)malloc(sizeof(struct RuleStk)*Nrule))==NULL){
		fprintf(stderr,"out of memory\n");
		exit(1);
	}
	memcpy(tmp,RList,sizeof(struct RuleStk)*Nrule);
	return(tmp);
}

struct AtomStk *SetCurrentAtoms(void)
{

	struct AtomStk *tmp;

	if((tmp=(struct AtomStk *)malloc(sizeof(struct AtomStk)*(Nvar+1)))==NULL){
		fprintf(stderr,"out of memory\n");
		exit(1);
	}
	memcpy(tmp,AList,sizeof(struct AtomStk)*(Nvar+1));
	return(tmp);
}

void InitializeCurrent(void)
{
	int i;

	if(Debug)
		fprintf(Errorfile, "InitializeCurrent\n");
	if((AList=(struct AtomStk *)malloc(sizeof(struct AtomStk)*(Nvar+1)))==NULL){
		fprintf(stderr,"out of memory\n");
		exit(1);
	}
	if((RList=(struct RuleStk *)malloc(sizeof(struct RuleStk)*Nrule))==NULL){
		fprintf(stderr,"out of memory\n");
		exit(1);
	}
	memset(RList,0,sizeof(struct RuleStk)*Nrule);
	for(i=0;i<Nrule;i++){
		RList[i].bdycnt=Rules[i].bdycnt;
		RList[i].hdcnt=Rules[i].hdcnt;
	}

	memset(AList,-1,sizeof(struct AtomStk)*(Nvar+1));
	for(i=1;i<=Nvar;i++){
		AList[i].pos=Atoms[i].pos;
		AList[i].neg=Atoms[i].neg;
		AList[i].cnt=Atoms[i].cnt;
	}
}	


void PushStck(int lit,int val)
{

	if(!BTtop)
		BTtop=NewStck();
	else if(!BTtop->up){
		BTtop->up=NewStck();
		BTtop->up->down=BTtop;
		BTtop=BTtop->up;
	}
	else BTtop=BTtop->up;
	BTtop->alst=SetCurrentAtoms();
	BTtop->rlst=SetCurrentRules();
	BTtop->atom=lit;
	BTtop->val=val;
	BTtop->satisfied=Sat;
	BTtop->assigned=Assigned;
}

int PopStck(int *val)
{
	int lit=0;

	if(!BTtop)
		return(0);
	free(RList);
	RList=BTtop->rlst;
	Sat=BTtop->satisfied;
	lit=BTtop->atom;
	(*val)=BTtop->val;
	Assigned=BTtop->assigned;
	free(AList);
	AList=BTtop->alst;
	BTtop=BTtop->down;
	return(lit);
}


void SetReference(void)
{

	int i,j,aptr;

	for(j=0;j<Nrule;j++){
		for(i=0;i<Rules[j].bdycnt;i++)
			Atoms[Rules[j].bdyptr[i]].bdycnt++;
		for(i=0;i<Rules[j].hdcnt;i++){
			Atoms[Rules[j].hdptr[i]].hdcnt++;
			if(Rules[j].horn)
				Atoms[Rules[j].hdptr[i]].horn=1;
		}
	}
	for(i=1;i<=Nvar;i++){
		if(Atoms[i].meta)
			for(j=0;j<Atoms[i].cnt;j++)
				Atoms[Atoms[i].metalist[j]].cnt++;
	}
	for(i=1;i<=Nvar;i++){
		Atoms[i].bdylst=NewData(sizeof(int)*(Atoms[i].bdycnt),0);
		Atoms[i].hdlst=NewData(sizeof(int)*(Atoms[i].hdcnt),0);
		if(!Atoms[i].meta){
			Atoms[i].metalist=NewData(sizeof(int)*(Atoms[i].cnt),0);
			Atoms[i].cnt=0;
		}
		Atoms[i].bdycnt=0;
		Atoms[i].hdcnt=0;
	}
	for(j=0;j<Nrule;j++){
		for(i=0;i<Rules[j].bdycnt;i++)
			Atoms[Rules[j].bdyptr[i]].bdylst[Atoms[Rules[j].bdyptr[i]].bdycnt++]=j;
		for(i=0;i<Rules[j].hdcnt;i++)
			Atoms[Rules[j].hdptr[i]].hdlst[Atoms[Rules[j].hdptr[i]].hdcnt++]=j;
		if(Rules[j].horn)
			Atoms[Rules[j].hdptr[0]].cnt++;
	}
	for(i=1;i<=Nvar;i++){
		if(Atoms[i].meta)
			for(j=0;j<Atoms[i].cnt;j++){
				aptr=Atoms[i].metalist[j];
				Atoms[aptr].metalist[Atoms[aptr].cnt++]=i;
			}
	}
}

void PrintTheory(void)
{
	int	i,j;

	fprintf(Errorfile," \n");
	for(i=0;i<Nrule;i++){
		fprintf(Errorfile,
				"Rule %i  %i: body count %i head count %i -- bodylist:",
				i,Rules[i].horn,Rules[i].bdycnt,Rules[i].hdcnt);
		for(j=0;j<Rules[i].bdycnt;j++)
			fprintf(Errorfile," %i ",Rules[i].bdyptr[j]);
		fprintf(Errorfile," -- headlist:");
		for(j=0;j<Rules[i].hdcnt;j++)
			fprintf(Errorfile," %i ",Rules[i].hdptr[j]);
		fprintf(Errorfile," \n");
	}

	fprintf(Errorfile," \n");
	for(i=1;i<=Nvar;i++){
		fprintf(Errorfile,"Atom %i %s value = %i -- ",
				i,Atoms[i].txt, AList[i].val);
			fprintf(Errorfile,"Atomlist:");
		if(Atoms[i].meta)
			fprintf(Errorfile,"MetaList:");
		for(j=0;j<Atoms[i].cnt;j++)
			fprintf(Errorfile," %i ",Atoms[i].metalist[j]);
		fprintf(Errorfile," \n");
	}
}



void RemoveRef(int rule)
{

	if(!RList[rule].sat){
		if(!Rules[rule].horn)
			Sat++;
		else AList[Rules[rule].hdptr[0]].cnt--;
	}
	RList[rule].sat=1;
}


int SetMeta(int ltl, int val)
{
	int setval=-1,i;
	 
	if(AList[ltl].req>=0&&AList[ltl].req!=val)
		return(0);
	if(Debug)
		fprintf(Errorfile, "atom = %i cnt = %i pos =%i neg= %i val=%i\n",
				ltl,AList[ltl].cnt,AList[ltl].pos,AList[ltl].neg,val);
	AList[ltl].req=val;
	if(val){
		if(AList[ltl].neg<=0&&AList[ltl].pos<=0)
			return(1);
		else if((AList[ltl].cnt<AList[ltl].pos)||(AList[ltl].cnt<AList[ltl].neg)||
			(AList[ltl].cnt<AList[ltl].pos+AList[ltl].neg))
			return(0);
		else if(AList[ltl].neg<=0&&AList[ltl].pos==AList[ltl].cnt)
			setval=1;
		else if(AList[ltl].pos<=0&&AList[ltl].neg==AList[ltl].cnt)
			setval=0;
	}
	else if((AList[ltl].cnt<AList[ltl].pos||AList[ltl].cnt<AList[ltl].neg)||
		(AList[ltl].cnt<AList[ltl].pos+AList[ltl].neg))
		return(1);
	else if(AList[ltl].cnt==1){
		if(AList[ltl].neg==1)
			setval=1;
		else if(AList[ltl].pos==1) setval=0;
	}
	if(setval>=0){
		for(i=0;i<Atoms[ltl].cnt;i++){
			if(AList[Atoms[ltl].metalist[i]].val==-1){
				if(AtomQueueVal[Atoms[ltl].metalist[i]]==-1){
					SetAtomQueue[SetAtomCnt++]=Atoms[ltl].metalist[i];
					AtomQueueVal[Atoms[ltl].metalist[i]]=setval;
				}
				else if(AtomQueueVal[Atoms[ltl].metalist[i]]!=setval)
					return(0);
			}
		}
	} 
	else return(-1);
	return(1);
} 

int SetHorn(int ltl,int horn,int val)
{
	 
	if(AList[ltl].req>=0&&AList[ltl].req!=val)
		return(0);
	AList[ltl].req=val;
	if(!horn) return(-1);
	return(1);
} 

int  PropAtoms(int *list,int cnt,int horn, int val)
{
	int i,set=-1,ret;

	for(i=0;i<cnt;i++){
		if(AList[list[i]].val<0){
			if(AtomQueueVal[list[i]]==-1){
				ret=1;
				if(Atoms[list[i]].horn){
					if((ret=SetHorn(list[i],horn,val))==0)
						return(0);
				}
				else if(Atoms[list[i]].meta){
					if((ret=SetMeta(list[i],val))==0)
						return(0);
				}
				if(ret==1){
					SetAtomQueue[SetAtomCnt++]=list[i];
					AtomQueueVal[list[i]]=val;
				}
				else if(ret==-1) set=1;
			}
			else if(AtomQueueVal[list[i]]!=val)
				return(0);
		}
	}
	return(set);
}

int ValSet(int v)
{
	if(v>0)
		return(1);
	else return(0);
}


int Opp(int v)
{
	if(v>0)
		return(0);
	else return(1);
}

int EvaluateRule(int rule);


int Closure(int rule,int level)
{
	int i,j,atom;

	if(Debug)
		fprintf(Errorfile, 
		"Closure level %i rule = %i atom = %i \n"
		,level, rule,Rules[rule].hdptr[0]);
	
	for(i=0;i<Rules[rule].bdycnt;i++){
		atom=Rules[rule].bdyptr[i];
		if((AList[atom].val<0)&&(AtomQueueVal[atom]<0)){
			if(Atoms[atom].horn){
				if(Debug)
					fprintf(Errorfile, "Pathlist = %i atom = %i \n"
					,PathList[atom], atom);
				if(PathList[atom])return(0);
				if(AList[atom].cnt==1){
					PathList[atom]=1;
					for(j=0;j<Atoms[atom].hdcnt;j++){
						if(Rules[Atoms[atom].hdlst[j]].horn&&
								!RList[Atoms[atom].hdlst[j]].sat){
							if(!Closure(Atoms[atom].hdlst[j],level+1)) return(0);
						}
					}
				}
				else if(!AList[atom].cnt) return(0);
			}
		}
	}
	for(i=0;i<Rules[rule].bdycnt;i++){
		atom=Rules[rule].bdyptr[i];
		if((AList[atom].val<0)&&(AtomQueueVal[atom]<0)&&!Atoms[atom].horn){
			SetAtomQueue[SetAtomCnt++]=atom;
			AtomQueueVal[atom]=1;
		}
	}
	return(1);
}

int CheckHorn(int atom)
{
	int i;
	if(Debug)
		fprintf(Errorfile, 
		"CheckHorn atom %i ",atom);
	if(AtomQueueVal[atom]>=0) return(1);

	if(AList[atom].cnt>1)
		return(1);
	else if(!AList[atom].cnt){
		if(AList[atom].req==1)
			return(0);
		AList[atom].req=0;
		SetAtomQueue[SetAtomCnt++]=atom;
		AtomQueueVal[atom]=0;
	}
	else if(AList[atom].cnt==1){
		if(AList[atom].req==1){
			memset(PathList,0,sizeof(short)*(Nvar+1));
			PathList[atom]=1;
			for(i=0;i<Atoms[atom].hdcnt;i++)
				if(Rules[Atoms[atom].hdlst[i]].horn&&!RList[Atoms[atom].hdlst[i]].sat)
					if(!Closure(Atoms[atom].hdlst[i],0)) return(0);
		}
	}
	return(1);
}

int PropRules(int atom,int val)
{
	int i,j,setval,atmp,rule;

	if(Debug)
		fprintf(Errorfile, 
		"atom = %i req= %i val = %i meta = %i cnt = %i\n",
		atom,AList[atom].req, val,Atoms[atom].meta,AList[atom].cnt);
	
	if(!Atoms[atom].meta&&!Atoms[atom].horn){
		for(i=0;i<Atoms[atom].cnt;i++){
			atmp=Atoms[atom].metalist[i];
			if(AList[atmp].val<0&&AtomQueueVal[atmp]<0){
				AList[atmp].cnt--;
				if(val)
					AList[atmp].pos--;
				else
					AList[atmp].neg--;
				if((AList[atmp].cnt<AList[atmp].pos)||
					(AList[atmp].cnt< AList[atmp].neg)||
					(AList[atmp].cnt<AList[atmp].pos+AList[atmp].neg))
					setval=0;
				else if(AList[atmp].pos<=0&&AList[atmp].neg<=0)
					setval=1;
				else setval=-1;
				if(setval>=0){
					if(AList[atmp].req>=0&&AList[atmp].req!=setval)
						return(0);
					SetAtomQueue[SetAtomCnt++]=atmp;
					AtomQueueVal[atmp]=setval;
				}
				else if(AList[atmp].req>=0){
					if(AList[atmp].req){
						if(AList[atmp].pos<=0&&AList[atmp].neg==AList[atmp].cnt)
							setval=0;
						else if(AList[atmp].neg<=0&&AList[atmp].pos==AList[atmp].cnt)
							setval=1;
					}
					else if(AList[atmp].cnt==1){
						if(AList[atmp].neg==1)
							setval=1;
						else if(AList[atmp].pos==1) 
							setval=0;
					}
					if(setval>=0){
						for(j=0;j<Atoms[atmp].cnt;j++){
							if(AList[Atoms[atmp].metalist[j]].val<0){
								if(AtomQueueVal[Atoms[atmp].metalist[j]]==-1){
									SetAtomQueue[SetAtomCnt++]=Atoms[atmp].metalist[j];
									AtomQueueVal[Atoms[atmp].metalist[j]]=setval;
								}
								else if(AtomQueueVal[Atoms[atmp].metalist[j]]!=setval)
									return(0);
							}
						}
						SetAtomQueue[SetAtomCnt++]=atmp;
						AtomQueueVal[atmp]=AList[atmp].req;
					}
				}
			}
		}
	}
	for(i=0;i<Atoms[atom].hdcnt;i++){
		rule=Atoms[atom].hdlst[i];
		if(!RList[rule].sat){
			if(val)
				RemoveRef(rule);
			else{
				RList[rule].hdcnt--;
				if(!EvaluateRule(rule))
					return(0);
				RList[rule].wgt*=Mult;
			}
		}
	}
	
	for(i=0;i<Atoms[atom].bdycnt;i++){
		rule=Atoms[atom].bdylst[i];
		if(!RList[rule].sat){
			if(val){
				RList[rule].bdycnt--;
				if(!EvaluateRule(rule))
					return(0);
				RList[rule].wgt*=Mult;
			}
			else{
				RemoveRef(rule);
				if(Rules[rule].horn){
					if(!CheckHorn(Rules[rule].hdptr[0]))
						return(0);
				}
			}
		}
	}
	return(1);
}


int SetAtom(int ltl, int val)
{
	 
	if(AList[ltl].val<0){
		AList[ltl].val=val;
		Assigned++;
		return(1);
	}
	else return(0);
} 


int Propagate(void)
{
	int i=0;

	while(i<SetAtomCnt){
		if(!SetAtom(SetAtomQueue[i], AtomQueueVal[SetAtomQueue[i]]))
			return(0);
		if(!PropRules(SetAtomQueue[i],AtomQueueVal[SetAtomQueue[i]]))
			return(0);
		i++;
	}
	if(Debug)
		fprintf(Errorfile, "Propagate cnt = %i\n",SetAtomCnt);
	if(Sat>=Nsat)
		return(-1);
	return(1);
}

int UnitProp(int literal, int val)
{
	SetAtomCnt=0;
	memset(AtomQueueVal,-1,sizeof(short)*(Nvar+1));
	SetAtomQueue[SetAtomCnt++]=literal;
	AtomQueueVal[literal]=val;
	return(Propagate());
}
int GetNextLit(int *literal)
{
	int lit=0,i,j,k,cnt,rule;
	double posmax=0,max=0,metamax=0;
	int mcnt=0,req=0,found=0,metamin=Nvar,min=Nvar;

	memset(Tmplist,0,sizeof(double)*(Nvar+1));
	memset(literal,0,sizeof(int)*(Nvar+1));
	for(i=1;i<=Outvar;i++){
		if(AList[i].val<0){
			if(!Atoms[i].horn){
				for(j=0;j<Atoms[i].hdcnt;j++)
					if(!RList[Atoms[i].hdlst[j]].sat){
						Tmplist[i]+=RList[Atoms[i].hdlst[j]].wgt;
					}
				for(j=0;j<Atoms[i].bdycnt;j++)
					if(!RList[Atoms[i].bdylst[j]].sat){
						Tmplist[i]+=RList[Atoms[i].bdylst[j]].wgt;
						if(Rules[Atoms[i].bdylst[j]].horn&&!found){
							rule=Atoms[i].bdylst[j];
							cnt=0;
							for(k=0;k<Rules[rule].bdycnt;k++)
								if((AList[Rules[rule].bdyptr[k]].val>0) ||
										(!Atoms[Rules[rule].bdyptr[k]].horn))
									cnt++;
							if(cnt==Rules[rule].bdycnt)
								found =1;
						}
					}
				for(j=0;j<Atoms[i].cnt;j++)
					if(AList[Atoms[i].metalist[j]].val<0){
						Tmplist[i]+=Weight[AList[Atoms[i].metalist[j]].cnt];
					}
				if(Tmplist[i]>max){
					max=Tmplist[i];
					lit = i;
				}
			}
			else if(AList[i].req) req=1;
		}
	}
	if(lit){
		for(i=Outvar+1;i<=Nvar;i++){
			if(AList[i].val<0){
				if(req&&!found){
					for(j=0;j<Atoms[i].bdycnt;j++)
						if((!RList[Atoms[i].bdylst[j]].sat)&&
							(Rules[Atoms[i].bdylst[j]].horn)) found =1;
				}
				if(AList[i].req==1&&AList[i].cnt<=metamin&&
						((AList[i].pos==1&&(1+AList[i].neg)==AList[i].cnt)
						 ||(AList[i].neg==1&&(1+AList[i].pos)==AList[i].cnt))){
					posmax=0;
					for(j=0;j<Atoms[i].cnt;j++)
						if(AList[Atoms[i].metalist[j]].val<0)
							posmax+=Tmplist[Atoms[i].metalist[j]];
					if(metamax<posmax||AList[i].cnt<metamin||Atoms[i].cnt<min){
						metamin=AList[i].cnt;
						metamax=posmax;
						min=Atoms[i].cnt;
						mcnt=0;
						for(j=0;j<Atoms[i].cnt;j++)
							if(AList[Atoms[i].metalist[j]].val<0){
								literal[mcnt]=Atoms[i].metalist[j];
								if(AList[i].pos>1) literal[mcnt]=-literal[mcnt];
								mcnt++;
							}
					}
				}
			}
		}
	}
	if(req && !found) return(0);
	if(mcnt){ 
		for(i=0;i<mcnt;i++){
			k=i;
			max=Tmplist[abs(literal[k])];
			for(j=i+1;j<mcnt;j++){
				if(Tmplist[abs(literal[j])]>max){
					max=Tmplist[abs(literal[j])];
					k=j;
				}
			}
			if(k!=i){
				lit=literal[i];
				literal[i]=literal[k];
				literal[k]=lit;
			}
		}
		return(mcnt);
	}
	else if(lit){
		literal[0]=lit;
		literal[1]=-lit;
		return(2);
	}
	return(0);
}


int Logic(void)
{
	int flag=1,more=1,i;
	int *literal,value=1;

	literal=NewData(sizeof(int)*(Nvar+1),0);
	while(more){	
		switch(flag){
		case 1:
			if(LAH)
				flag=LookAHead(literal);
			else
				flag=GetNextLit(literal);
			if(flag>0){
				Branch++;
				for(i=1;i<flag;i++){
					if(Debug)
						fprintf(Errorfile, "flag=%i literal[%i]=%i  \n",
								flag,i,literal[i]); 
					PushStck(abs(literal[i]),ValSet(literal[i]));
				}
				flag=UnitProp(abs(literal[0]),ValSet(literal[0]));
			}
			if(Debug){
				fprintf(Errorfile,
				"flag=%i branch=%i literal=%i  Assign=%i Sat=%i\n", 
				flag,Branch,literal[0],Assigned,Sat);
				getchar();
			}
			break;
		case 0:
			Backtrack++;
			if((literal[0]=PopStck(&value))){
				flag=UnitProp(literal[0],value);
				if(Debug){
					fprintf(Errorfile,
					"flag=%i Backtrack=%i literal=%i value =%i Assign=%i Sat=%i\n", 
					flag,Backtrack,literal[0],value,Assigned,Sat);
					getchar();
				}
			}
			else{
				if(Count&&Cnt)
					flag=-1;
				more=0;
			}
			break;
		case -1: 
			if(Answer)
				PrintAnswer();
			if(Count){
				SetCnt();
				if(Count>0&&Count<Cnt)
					more=0;
				else flag=1;
			}
			else{
			       Cnt=1;
			       more=0;
			}
			break;
		default :
			more=0;
			break;
		}
	}
	return(flag);
}



int EvaluateRule(int rule)
{
	int ret=1;

	if(RList[rule].sat)
		return(1);
	if(!RList[rule].bdycnt){
		if(!RList[rule].hdcnt)
			return(0);
		else if(RList[rule].hdcnt==1){
			if((ret=PropAtoms(Rules[rule].hdptr,Rules[rule].hdcnt,Rules[rule].horn,1))==-1){ 
				ret=1;
				RemoveRef(rule);
			}
		}
	}
	else if(RList[rule].bdycnt==1&&!RList[rule].hdcnt){
		if((ret=PropAtoms(Rules[rule].bdyptr,Rules[rule].bdycnt,0,0))==-1){
			ret=1;
			RemoveRef(rule);
		}
	}

	return(ret);
}



int PreProcess(void)
{
	int i;

	memset(AtomQueueVal,-1,sizeof(short)*(Nvar+1));
	SetAtomCnt=0;
	for(i=0;i<Nrule;i++){
		if(!EvaluateRule(i))
			return(0);
	}
	if(SetAtomCnt)
		if(!Propagate())
			return(0);
	memset(AtomQueueVal,-1,sizeof(short)*(Nvar+1));
	SetAtomCnt=0;
	for(i=1;i<=Outvar;i++){
		if(Atoms[i].horn&&AList[i].req==1&&AList[i].val<0&&AtomQueueVal[i]<0){
			if(!CheckHorn(i)) return(0);
		}
	}
	if(SetAtomCnt)
		if(!Propagate())
			return(0);
	if(Nsat<=Sat){
		Cnt++;
		if(Answer)
			PrintAnswer();
		return(-1);
	}
	return(1);
}

void SetWeights(void)
{
	int max=10,i;
	double	wgt;

	Weight=(double *)NewData(sizeof(double)*(Maxlit+1),0);
	wgt=1;
	for(i=Maxlit;i>max;i--)
		Weight[i]=wgt;
	if(Maxlit<max) max=Maxlit;
	for(i=max;i>=0;i--){
		Weight[i]=wgt;
		wgt*=Mult;
	}
	for(i=0;i<Nrule;i++){
		RList[i].wgt=Weight[RList[i].bdycnt+RList[i].hdcnt];
	}
	if(Debug){
		for(i=0;i<=Maxlit;i++)
			fprintf(Errorfile," Weight[%i]=%g",i,Weight[i]);
		fprintf(Errorfile,"\n");
	}
}

void InitializeLookAHead(void);

int SetUp(char *filename)
{

	FILE 		*summary;
	static char	*solution[]={"unsat","sat","unknown","error"};
	int			answer=0,sol=0;
	struct tms 	cpu;
	struct stat buf;
	double 		cputime=0,begin=0;


	if((sol=GetProblem())==-1)
		return(sol);
	if(sol){
		SetReference();
		InitializeCurrent();

		SetWeights();

		if(Print){
			PrintTheory();
			return(0);
		}

		if(Restart<0||(Incomplete&&!Restart))
			Restart=Nvar;
		times(&cpu);
		begin=cpu.tms_utime;
		answer=0;
		sol=PreProcess();
		while(sol==1){
	/*
		if(Restart){
			if(Debug){
				fprintf(stderr,"Restart = %i ",Restart);
				fprintf(stderr,"Assigned = %i Sat = %i \n",Assigned,Sat);
			}
			StackStore(1,0);
			Assigned=Stcklist[1]=Stcklist[0];
			Sat=Satlist[1]=Satlist[0];
			sol=Logic(&answer,1);
		}
		else if(!Incomplete){
		*/
			sol=Logic();
	/*
		}
		else sol=2;
		*/
		}
	}
	if(Debug)
		fprintf(stderr,"sol = %i ",sol);
	times(&cpu);
	cputime=(cpu.tms_utime-begin)/(double)CLK_TCK;
	if(Debug)
		fprintf(stderr,"cputime = %g ",cputime);
	if(sol==-1)
	/*{
		sol=CheckSol(Valstk[answer]);
		Cnt=sol;
	}
	else if(Count&&Cnt)
	*/
		sol=1;
	/*
	answer=0;
	*/
	if(stat("aspps.stat",&buf)==-1)
		answer=1;
	if((summary=fopen("aspps.stat","a"))==NULL){
		fprintf(Errorfile, "cannot open aspps.stat");
		return(0);
	}
	if(answer)
		fprintf(summary,"Variable  Clauses   Sat   CPU Time  Branch    Bcktrck  Count     Filename\n");
	fprintf(summary,"%-10i%-10i%-6s%-10g%-10i%-10i%-10i",Nvar,Nrule,solution[sol],cputime,Branch,Backtrack,Cnt);
	fprintf(summary,"%s\n",filename);
	fclose(summary);
	return(1);
}

	

int main(int argc, char **argv){

	char	filename[130];
	int	i,j,fn=0;

	Print=0;
	Restart=0;
	Incomplete=0;
	NonRand=1;
	Sat=0;
	Count=0;
	Cnt=0;
	Local=0;
	Answer=0;
	Backtrack=0;
	CQ=0;
	CQlit=0;
	srand48((unsigned)time(NULL));
	getcwd(filename,129); 
	ShowAtoms=NULL;
	printf("\n%s\n",VERSION);
	i=1;
	while(i<argc){
		if(argv[i][0]=='-')
			switch(argv[i][1]){
				case 'P': Print=1;break; 
				case 'f': fn=1;strcpy(filename,argv[++i]);break;
				case 'C': if((i+1==argc)||(argv[i+1][0]=='-')) 
							  Count=-1;
						  else Count=atoi(argv[++i]);
						  break;
				case 'S': while(ShowCnt+i+1<argc&&argv[i+ShowCnt+1][0]!='-') ShowCnt++;
						  if((ShowAtoms=(char **)malloc(sizeof(char *)*ShowCnt))==NULL)
							  exit(1);
						  j=0;
						  while(j<ShowCnt)
							  ShowAtoms[j++]=strdup(argv[++i]);
						  if(ShowCnt)
							  Answer=1;
						  break;
				case 'L': if((i+1==argc)||(argv[i+1][0]=='-')) 
							  LAH=5;
						  else
							  LAH=atoi(argv[++i]);
						  break;
				case 'A': Answer=1;break;
				case 'D': Debug=1;break;
				default: fprintf(stderr,
					 "\nuseage: aspps [-f filename] [-L [nn] ] [-C [nn]] [-A] [-S predicate] \n");
						fprintf(stderr, "-L : Adds lookahead with default of 5 atoms \n");
						fprintf(stderr, "-L nn : Adds lookahead with nn atoms \n");
						fprintf(stderr, "-C : finds all solutions.  Number of solutions written to stat file \n");
						fprintf(stderr, "-C nn: finds first nn solutions.\n");
						fprintf(stderr, "-A : Displays all positive atoms\n");
						fprintf(stderr, 
								"-S predicate : Displays all positive atoms constructed from predicate\n");
				  return(1);
		}
		i++;
	}
	Errorfile=stderr;
	if(!fn){
		fptr = stdin;
		SetUp("stdin");
	}
	else if((fptr=fopen(filename,"r"))==NULL){
		printf("error: can not open  %s \n",filename);
		return(-1);
	}
	else 
		SetUp(filename);
	return(0);
}


int LookAHead(int *literal )
{
	int 	lit=0,more=1,pflag;
	int		i,val,lcnt=0,scnt=0,cnt=0;

	if(Debug)
		fprintf(stderr,"LookAHead \n");
	more=LAH;
	while(more>0){
		if((cnt=GetNextLit(literal))&&cnt<3){
			scnt=0;
			for(i=0;i<cnt;i++){
				more--;
				PushStck(abs(literal[i]),ValSet(literal[i]));
				pflag=UnitProp(abs(literal[i]),ValSet(literal[i]));
				if(pflag==-1) return(pflag);
				if(pflag) lit=i;
				scnt+=pflag;
				PopStck(&val);
			}
			if(!scnt)return(0);
			else if(scnt==1){
				UnitProp(abs(literal[lit]),ValSet(literal[lit]));
				more=LAH;
				lcnt++;
			}
		}
		else return(cnt);
	}

	if(Debug)
		fprintf(stderr,"LookAHead lit=%i lcnt=%i\n",literal[0],lcnt);
	return(cnt);
}


